@@ -1,22 +1,16 @@
+from __future__ import annotations
+
 from datetime import datetime
 from datetime import timedelta
 import functools
-import json
 import logging
 import os
 import threading
-import traceback
 import typing
 from typing import Any
-from typing import Callable
-from typing import cast
-from typing import Dict
-from typing import List
 from typing import Optional
-from typing import TypeVar
 from typing import Union
 
-from bottle import BaseResponse
 from bottle import Bottle
 from bottle import redirect
 from bottle import request
@@ -24,32 +18,39 @@ from bottle import response
 from bottle import run
 from bottle import SimpleTemplate
 from bottle import static_file
+import optuna
 from optuna.exceptions import DuplicatedStudyError
 from optuna.storages import BaseStorage
 from optuna.storages import RDBStorage
 from optuna.study import StudyDirection
 from optuna.study import StudySummary
 from optuna.trial import FrozenTrial
+from optuna.trial import TrialState
 from optuna.version import __version__ as optuna_ver
 from packaging import version
 
 from . import _note as note
+from ._bottle_util import BottleViewReturn
+from ._bottle_util import json_api_view
 from ._cached_extra_study_property import get_cached_extra_study_property
 from ._importance import get_param_importance_from_trials_cache
+from ._pareto_front import get_pareto_front_trials
 from ._serializer import serialize_study_detail
 from ._serializer import serialize_study_summary
+from ._storage_url import get_storage
+from .artifact._backend import delete_all_artifacts
+from .artifact._backend import register_artifact_route
 
 
 if typing.TYPE_CHECKING:
     from _typeshed.wsgi import WSGIApplication
+    from optuna_dashboard.artifact.protocol import ArtifactBackend
 
     try:
         from optuna.study._frozen import FrozenStudy
     except ImportError:
         FrozenStudy = None  # type: ignore
 
-BottleViewReturn = Union[str, bytes, Dict[str, Any], BaseResponse]
-BottleView = TypeVar("BottleView", bound=Callable[..., BottleViewReturn])
 
 logger = logging.getLogger(__name__)
 
@@ -61,8 +62,8 @@ cached_path_exists = functools.lru_cache(maxsize=10)(os.path.exists)
 
 # In-memory trials cache
 trials_cache_lock = threading.Lock()
-trials_cache: Dict[int, List[FrozenTrial]] = {}
-trials_last_fetched_at: Dict[int, datetime] = {}
+trials_cache: dict[int, list[FrozenTrial]] = {}
+trials_last_fetched_at: dict[int, datetime] = {}
 
 # RDB schema migration check
 rdb_schema_migrate_lock = threading.Lock()
@@ -123,26 +124,11 @@ def update_schema_compatibility_flags(storage: BaseStorage) -> None:
         rdb_schema_unsupported = current_version not in storage.get_all_versions()
 
 
-def json_api_view(view: BottleView) -> BottleView:
-    @functools.wraps(view)
-    def decorated(*args: List[Any], **kwargs: Dict[str, Any]) -> BottleViewReturn:
-        try:
-            response.content_type = "application/json"
-            response_body = view(*args, **kwargs)
-            return response_body
-        except Exception as e:
-            response.status = 500
-            response.content_type = "application/json"
-            stacktrace = "\n".join(traceback.format_tb(e.__traceback__))
-            logger.error(f"Exception: {e}\n{stacktrace}")
-            return json.dumps({"reason": "internal server error"})
-
-    return cast(BottleView, decorated)
-
-
-def get_study_summaries(storage: BaseStorage) -> List[StudySummary]:
+def get_study_summaries(storage: BaseStorage) -> list[StudySummary]:
     if version.parse(optuna_ver) >= version.Version("3.0.0rc0.dev"):
         frozen_studies = storage.get_all_studies()  # type: ignore
+        if isinstance(storage, RDBStorage):
+            frozen_studies = sorted(frozen_studies, key=lambda s: s._study_id)
         return [_frozen_study_to_study_summary(s) for s in frozen_studies]
     elif version.parse(optuna_ver) >= version.Version("3.0.0b0.dev"):
         return storage.get_all_study_summaries(include_best_trial=False)  # type: ignore
@@ -160,7 +146,7 @@ def get_study_summary(storage: BaseStorage, study_id: int) -> Optional[StudySumm
 
 
 def create_new_study(
-    storage: BaseStorage, study_name: str, directions: List[StudyDirection]
+    storage: BaseStorage, study_name: str, directions: list[StudyDirection]
 ) -> int:
     if version.parse(optuna_ver) >= version.Version("3.1.0.dev") and version.parse(
         optuna_ver
@@ -172,7 +158,7 @@ def create_new_study(
     return study_id
 
 
-def get_trials(storage: BaseStorage, study_id: int, ttl_seconds: int = 10) -> List[FrozenTrial]:
+def get_trials(storage: BaseStorage, study_id: int, ttl_seconds: int = 10) -> list[FrozenTrial]:
     with trials_cache_lock:
         trials = trials_cache.get(study_id, None)
         last_fetched_at = trials_last_fetched_at.get(study_id, None)
@@ -198,7 +184,11 @@ def get_trials(storage: BaseStorage, study_id: int, ttl_seconds: int = 10) -> Li
     return trials
 
 
-def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
+def create_app(
+    storage: BaseStorage,
+    artifact_backend: Optional[ArtifactBackend] = None,
+    debug: bool = False,
+) -> Bottle:
     app = Bottle()
     update_schema_compatibility_flags(storage)
 
@@ -241,9 +231,16 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
             rdb_schema_needs_migrate = False
         return redirect("/dashboard", 302)
 
+    @app.get("/api/meta")
+    @json_api_view
+    def api_meta() -> dict[str, Any]:
+        return {
+            "artifact_is_available": artifact_backend is not None,
+        }
+
     @app.get("/api/studies")
     @json_api_view
-    def list_study_summaries() -> BottleViewReturn:
+    def list_study_summaries() -> dict[str, Any]:
         summaries = get_study_summaries(storage)
         serialized = [serialize_study_summary(summary) for summary in summaries]
         return {
@@ -252,7 +249,7 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
 
     @app.post("/api/studies")
     @json_api_view
-    def create_study() -> BottleViewReturn:
+    def create_study() -> dict[str, Any]:
         study_name = request.json.get("study_name", None)
         request_directions = [d.lower() for d in request.json.get("directions", [])]
         if (
@@ -280,20 +277,59 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
         response.status = 201  # Created
         return {"study_summary": serialize_study_summary(summary)}
 
+    @app.post("/api/studies/<study_id:int>/rename")
+    @json_api_view
+    def rename_study(study_id: int) -> dict[str, Any]:
+        dst_study_name = request.json.get("study_name", None)
+        if dst_study_name is None:
+            response.status = 400  # Bad request
+            return {"reason": "You need to set study_name and direction"}
+
+        src_study_name = storage.get_study_name_from_id(study_id)
+        try:
+            src_study = optuna.load_study(storage=storage, study_name=src_study_name)
+        except KeyError:
+            response.status = 404  # Not found
+            return {"reason": f"study_id={study_id} is not found"}
+
+        try:
+            dst_study = optuna.create_study(storage=storage, study_name=dst_study_name)
+            dst_study.add_trials(src_study.get_trials(deepcopy=False))
+        except DuplicatedStudyError:
+            response.status = 400  # Bad request
+            return {"reason": f"study_name={dst_study_name} is duplicaated"}
+        except Exception as e:
+            logger.exception("Unexpected error:")
+            response.status = 500
+            storage.delete_study(dst_study._study_id)
+            return {"reason": str(e)}
+        new_study_summary = get_study_summary(storage, dst_study._study_id)
+        if new_study_summary is None:
+            response.status = 500
+            return {"reason": "Failed to load the new study"}
+
+        storage.delete_study(src_study._study_id)
+        response.status = 201
+        return serialize_study_summary(new_study_summary)
+
     @app.delete("/api/studies/<study_id:int>")
     @json_api_view
-    def delete_study(study_id: int) -> BottleViewReturn:
+    def delete_study(study_id: int) -> dict[str, Any]:
+        if artifact_backend is not None:
+            system_attrs = storage.get_study_system_attrs(study_id)
+            delete_all_artifacts(artifact_backend, system_attrs)
+
         try:
             storage.delete_study(study_id)
         except KeyError:
             response.status = 404  # Not found
             return {"reason": f"study_id={study_id} is not found"}
         response.status = 204  # No content
-        return ""
+        return {}
 
     @app.get("/api/studies/<study_id:int>")
     @json_api_view
-    def get_study_detail(study_id: int) -> BottleViewReturn:
+    def get_study_detail(study_id: int) -> dict[str, Any]:
         try:
             after = int(request.params["after"])
             assert after >= 0
@@ -307,6 +343,12 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
             response.status = 404  # Not found
             return {"reason": f"study_id={study_id} is not found"}
         trials = get_trials(storage, study_id)
+
+        # TODO(c-bata): Cache best_trials
+        if summary.directions == 1:
+            best_trials = [storage.get_best_trial(study_id)]
+        else:
+            best_trials = get_pareto_front_trials(trials=trials, directions=summary.directions)
         (
             # TODO: intersection_search_space and union_search_space look more clear since now we
             # have union_user_attrs.
@@ -317,6 +359,7 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
         ) = get_cached_extra_study_property(study_id, trials)
         return serialize_study_detail(
             summary,
+            best_trials,
             trials[after:],
             intersection,
             union,
@@ -326,28 +369,27 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
 
     @app.get("/api/studies/<study_id:int>/param_importances")
     @json_api_view
-    def get_param_importances(study_id: int) -> BottleViewReturn:
-        # TODO(chenghuzi): add support for selecting params via query parameters.
-        objective_id = int(request.params.get("objective_id", 0))
+    def get_param_importances(study_id: int) -> dict[str, Any]:
         try:
             n_directions = len(storage.get_study_directions(study_id))
         except KeyError:
             response.status = 404  # Study is not found
             return {"reason": f"study_id={study_id} is not found"}
-        if objective_id >= n_directions:
-            response.status = 400  # Bad request
-            return {"reason": f"study_id={study_id} has only {n_directions} direction(s)."}
 
         trials = get_trials(storage, study_id)
         try:
-            return get_param_importance_from_trials_cache(storage, study_id, objective_id, trials)
+            importances = [
+                get_param_importance_from_trials_cache(storage, study_id, objective_id, trials)
+                for objective_id in range(n_directions)
+            ]
+            return {"param_importances": importances}
         except ValueError as e:
             response.status = 400  # Bad request
             return {"reason": str(e)}
 
     @app.put("/api/studies/<study_id:int>/note")
     @json_api_view
-    def save_note(study_id: int) -> BottleViewReturn:
+    def save_study_note(study_id: int) -> dict[str, Any]:
         req_note_ver = request.json.get("version", None)
         req_note_body = request.json.get("body", None)
         if req_note_ver is None or req_note_body is None:
@@ -355,15 +397,76 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
             return {"reason": "Invalid request."}
 
         system_attrs = storage.get_study_system_attrs(study_id)
-        if not note.version_is_incremented(system_attrs, req_note_ver):
+        if not note.version_is_incremented(system_attrs, None, req_note_ver):
             response.status = 409  # Conflict
             return {
                 "reason": "The text you are editing has changed. "
                 "Please copy your edits and refresh the page.",
-                "note": note.get_note_from_system_attrs(system_attrs),
+                "note": note.get_note_from_system_attrs(system_attrs, None),
             }
 
-        note.save_note(storage, study_id, req_note_ver, req_note_body)
+        note.save_note_with_version(storage, study_id, None, req_note_ver, req_note_body)
+        response.status = 204  # No content
+        return {}
+
+    @app.post("/api/trials/<trial_id:int>/tell")
+    @json_api_view
+    def tell_trial(trial_id: int) -> dict[str, Any]:
+        if "state" not in request.json:
+            response.status = 400  # Bad request
+            return {"reason": "state must be specified."}
+
+        try:
+            state = TrialState[request.json["state"].upper()]
+        except Exception:  # To catch KeyError and Exception by non str case.
+            response.status = 400  # Bad request
+            return {"reason": "state must be either 'Complete' or 'Fail'."}
+
+        if state not in [TrialState.COMPLETE, TrialState.FAIL]:
+            response.status = 400  # Bad request
+            return {"reason": "state must be either 'Complete' or 'Fail'."}
+
+        values = None
+        if state == TrialState.COMPLETE:
+            vs = request.json.get("values")
+            if vs is None:
+                response.status = 400  # Bad request
+                return {"reason": "values attribute is required when state is 'Complete'."}
+            try:
+                values = [float(v) for v in vs]
+            except (ValueError, TypeError):
+                response.status = 400  # Bad request
+                return {"reason": "values attribute must be an array of numbers"}
+
+        try:
+            storage.set_trial_state_values(trial_id, state, values)
+        except Exception as e:
+            response.status = 500
+            return {"reason": f"Internal server error: {e}"}
+
+        response.status = 204
+        return {}
+
+    @app.put("/api/studies/<study_id:int>/<trial_id:int>/note")
+    @json_api_view
+    def save_trial_note(study_id: int, trial_id: int) -> dict[str, Any]:
+        req_note_ver = request.json.get("version", None)
+        req_note_body = request.json.get("body", None)
+        if req_note_ver is None or req_note_body is None:
+            response.status = 400  # Bad request
+            return {"reason": "Invalid request."}
+
+        # Store note content in study system attrs since it's always updatable.
+        system_attrs = storage.get_study_system_attrs(study_id=study_id)
+        if not note.version_is_incremented(system_attrs, trial_id, req_note_ver):
+            response.status = 409  # Conflict
+            return {
+                "reason": "The text you are editing has changed. "
+                "Please copy your edits and refresh the page.",
+                "note": note.get_note_from_system_attrs(system_attrs, trial_id),
+            }
+
+        note.save_note_with_version(storage, study_id, trial_id, req_note_ver, req_note_body)
         response.status = 204  # No content
         return {}
 
@@ -381,6 +484,7 @@ def create_app(storage: BaseStorage, debug: bool = False) -> Bottle:
                 filename = gz_filename
         return static_file(filename, root=STATIC_DIR)
 
+    register_artifact_route(app, storage, artifact_backend)
     return app
 
 
@@ -400,31 +504,29 @@ def _frozen_study_to_study_summary(frozen_study: "FrozenStudy") -> StudySummary:
     )
 
 
-def get_storage(storage: Union[str, BaseStorage]) -> BaseStorage:
-    if isinstance(storage, str):
-        if storage.startswith("redis"):
-            raise ValueError(
-                "RedisStorage is unsupported from Optuna v3.1 or Optuna Dashboard v0.8.0"
-            )
-        else:
-            return RDBStorage(storage, skip_compatibility_check=True)
-    return storage
-
-
 def run_server(
-    storage: Union[str, BaseStorage], host: str = "localhost", port: int = 8080
+    storage: Union[str, BaseStorage],
+    host: str = "localhost",
+    port: int = 8080,
+    artifact_backend: Optional[ArtifactBackend] = None,
 ) -> None:
     """Start running optuna-dashboard and blocks until the server terminates.
+
     This function uses wsgiref module which is not intended for the production
     use. If you want to run optuna-dashboard more secure and/or more fast,
-    please use WSGI server like Gunicorn or uWSGI via `wsgi()` function.
+    please use WSGI server like Gunicorn or uWSGI via :func:`wsgi` function.
+
+
     """
-    app = create_app(get_storage(storage))
+    app = create_app(get_storage(storage), artifact_backend=artifact_backend)
     run(app, host=host, port=port)
 
 
-def wsgi(storage: Union[str, BaseStorage]) -> "WSGIApplication":
+def wsgi(
+    storage: Union[str, BaseStorage],
+    artifact_backend: Optional[ArtifactBackend] = None,
+) -> WSGIApplication:
     """This function exposes WSGI interface for people who want to run on the
     production-class WSGI servers like Gunicorn or uWSGI.
     """
-    return create_app(get_storage(storage))
+    return create_app(get_storage(storage), artifact_backend=artifact_backend)
